%This file contains the code to run the nascent/mature RNA Gillespie
%simulation.
%To run correctly, it requires a driver and an auxiliary definition file
%for class HandleState_int.

%gillespie172 converts units and initializes some hidden values to
%defaults. It calls gill and reports the results.
%gill initializes the simulation's data structures and calls the 
%function gill_timestepping with different settings to represent
%equilibration, delay, and normal operation.
%gill_timestepping iterates through the Gillespie algorithm loop, uses the
%function rxn_calculator to select reactions, uses probe_calculator to
%calculate fluorescence outputs, and uses markovstep to edit the data
%structures. When the simulation halts, it returns last state and probe
%values.
%probe_calculator predicts probe fluorescence values for each molecule.
%rxn_calculator calculates propensities and uses them to increment time and
%select the next reaction for each cell.
%markovstep modifies the system state.
%find_glob_index chooses an mRNA to undergo a reaction and reports its
%index in the dynamic data structure.

%gillespie171 is a wrapper for the Gillespie function. It performs unit conversions
%and initializes missing values to their defaults.
%Inputs: time_array - array of times at which to evaluate the simulation
%                     state. Must have as many rows as there are cells to
%                     evaluate.
%        p_phys_stoch - structure with physical properties of the system.
%                       Has gene length (L), number of gene copies
%                       (ngenes), probe indices (probes); optionally,
%                       delay length (delay).
%        p_sim_stoch - structure with simulation properties. 
%                      Has number of cells to evaluate (nCells), simulation
%                      coarseness in nt per elongation step (Nuc);
%                      optionally, equilibration time (Teq) and initial
%                      fraction of genes in ON state (initon_fraction).
%        units - string defining units of time_array and kin structures.
%                Can be 's', 'min', or 'h'; the internal calculations are
%                in minutes.
%        experiment_mode - string defining type of experiment. Can be 'on',
%                          'off', or 'ss' for turnon, turnoff, and
%                          steady-state respectively. 'off' and 'ss'
%                          require equilibration.
%        kin - array defining reaction rates. Must have as many rows as
%              there are cells to evaluate and units consistent with
%              time_array. The five columns correspond to
%              [kinit kon koff kd v], with v in units of nt/time and all
%              others in units of 1/time
%Outputs: nas_mesh - 4-dimensional array of probe signal outputs from
%                    nascent mRNA. Dimensions are in order of cell index,
%                    time point, probe index, gene copy index.
%         cyt_mesh - 3-dimenisional array of probe signal outputs from
%                    mature RNA. Dimensions are in order of cell index,
%                    time point, probe index.
%         tot_mesh - 3-dimensional array of total probe signal outputs.
%                    Dimensions are in order of cell index, time point, 
%                    probe index.
%         debug_mesh - 3-dimensional array of molecule numbers, which are
%                      not directly observed. Can be used to evaluate
%                      average molecule signal intensity, or modified
%                      altogether to report different variables for debug
%                      purposes. Dimensions are in order of cell index,
%                      time point, population (nascent, then mature).
function [nas_mesh,cyt_mesh,tot_mesh,debug_mesh,movie_struct] = ...
    gillespie172(time_array,p_phys_stoch,p_sim_stoch,units,experiment_mode,kin)
%Extraction of simulation and physical parameters.
nCells = p_sim_stoch.Niter; %number of iterations
Nuc = p_sim_stoch.Nuc; %nucleotides per step of elongation reaction
gene_length = p_phys_stoch.L;
ngenes = p_phys_stoch.ngenes;
% disp('pass into 172')

%Unit conversions. The internal script timer is in minutes. 
if strcmp(units,'s')
    unit_conv_const = 60;
elseif strcmp(units,'min')
    unit_conv_const = 1;
elseif strcmp(units,'h')
    unit_conv_const = 1/60;
else 
    error('Please specify parameter units as ''s'', ''h'', or ''min''.');
end

k = NaN(nCells,8);
k(:,1) = unit_conv_const*kin(:,1); %initiation
k(:,2) = unit_conv_const*kin(:,5)/Nuc; %elongation
k(:,3) = unit_conv_const*kin(:,4); %attachment of RNAse to nascent RNA
k(:,7) = unit_conv_const*kin(:,2); %turnon
k(:,8) = unit_conv_const*kin(:,3); %turnoff

k(:,4) = k(:,3); %attachment of RNAse to mature RNA
k(:,5) = k(:,2); %degradation of nascent RNA
k(:,6) = k(:,2); %degradation of mature RNA

% size(time_array)
ts_min = time_array/unit_conv_const;

probes = p_phys_stoch.probes*gene_length; %find RNA indices corresponding 
%                                          to probe ends


%Converting or filling in delay
if isfield(p_phys_stoch,'delay')
    delay = p_phys_stoch.delay/unit_conv_const;
else
    delay = 0;
end

%Converting or filling in equilibration time
if isfield(p_sim_stoch,'Teq')
    Teq = p_sim_stoch.Teq/unit_conv_const;
    if strcmp(experiment_mode,'on')
        warning('Equilibration time set for turnon experiment.');
    end
elseif strcmp(experiment_mode,'off') || strcmp(experiment_mode,'ss')
    warning('Equilibration time set to default value');
    
    %Need to remove rates of zero that would force Teq to be infinite.
    %This is a heuristic.
    k_eq = k; k_eq(k_eq<eps) = NaN; 
    Teq = 5./min(k_eq,[],2);
elseif strcmp(experiment_mode,'on')
    Teq = 0;
end

%Initializing the initial fraction of genes that are on.s
if isfield(p_sim_stoch,'initon_fraction')
    initon_fraction = p_sim_stoch.initon_fraction;
else
    initon_fraction = 0;
end

num_t_pts = size(ts_min,2);

%store reaction parameters
rxn_params = struct; 
rxn_params.k = k; rxn_params.gene_length = gene_length;
rxn_params.Nuc = Nuc; rxn_params.num_probes = size(probes,2); rxn_params.nCells = nCells;
rxn_params.probes = probes; 
rxn_params.ngenes = ngenes;

%check whether the code needs to store info for animation
% animation_mode = nargin>6;
animation_mode = isfield(p_sim_stoch,'animate') && p_sim_stoch.animate;

[nascentprobemesh,matureprobemesh,debug_mesh,movie_struct] = ...
    gill(rxn_params,experiment_mode,delay,ts_min,initon_fraction,Teq,animation_mode);

%Converting from nt to normalized probe signals
probediff_nas = repmat(reshape(diff(probes),[1,1,size(probes,2)]),...
    [nCells,num_t_pts,1,ngenes]);
probediff_cyt = repmat(reshape(diff(probes),[1,1,size(probes,2)]),...
    [nCells,num_t_pts,1]);
nas_mesh = nascentprobemesh ./ probediff_nas; %ONLY ONE PROBE FOR NOW!
cyt_mesh = matureprobemesh ./ probediff_cyt;
tot_mesh = sum(nas_mesh,4) + cyt_mesh;
return

%gill initializes data structures and runs the possibly heterogeneous
%sequence of parameter settings (equilibration, delay, normal operation).
%Inputs: k - rates of all possible reactions in system, in 1/min, as
%            defined in gillespie171 wrapper.
%        Nuc - simulation coarseness in nt/step of elongation. Must be an
%              integer.
%        gene_length - gene length in nt.
%        probes - probe indices in nt. The two rows are the 5' and 3' end,
%                 the columns are distinct probes.
%        experiment_mode - same as in gillespie171.
%        delay - delay in min.
%        time_array - array of times to sample system at, in min. Otherwise
%                     same as in gillespie171.
%        nCells - number of cells evaluated in the simulation.
%        initon_fraction - fraction of genes in ON state at the beginning
%                          of a turnon experiment.
%        Teq - equilibration time in min; not relevant to turnon.
%Outputs: nascentprobemesh - 4-dimensional array of probe signal outputs from
%                    nascent mRNA. Dimensions are in order of cell index,
%                    time point, probe index, gene copy index. Not
%                    normalized to gene length.
%         matureprobemesh - 3-dimenisional array of probe signal outputs from
%                    mature RNA. Dimensions are in order of cell index,
%                    time point, probe index. Not normalized to gene
%                    length.
%         debug_mesh - Same as in gillespie171 outputs.
function [nascentprobemesh,matureprobemesh,debug_mesh,movie_struct] = ...
    gill(rxn_params,experiment_mode,delay,time_array,...
    initon_fraction,Teq,animation_mode)
%extract relevant variables
num_probes = size(rxn_params.probes,2);
num_t_pts = size(time_array,2); %must be a vector
nCells = rxn_params.nCells;
ngenes = rxn_params.ngenes;

%initialize probe value and test arrays
nascentprobemesh = NaN(nCells,num_t_pts,num_probes,ngenes); 
matureprobemesh = NaN(nCells,num_t_pts,num_probes);
debug_mesh = zeros(nCells,num_t_pts,ngenes+1); 

%initialize structure for animated visualization
if animation_mode
    movie_struct = struct;
    movie_struct.RNA_53_indices = cell(num_t_pts,1);
    movie_struct.RNA_identity = cell(num_t_pts,1);
    movie_struct.RNA_rnase_binding = cell(num_t_pts,1);
    movie_struct.fill = cell(num_t_pts,1);
    movie_struct.x = cell(num_t_pts,1);
    movie_struct.rnase = cell(num_t_pts,1);
    movie_struct.gene_state = cell(num_t_pts,1);
else
    movie_struct = [];
end

%initalize all gene copies
geneactivestate = rand(nCells,ngenes);
geneactivestate = geneactivestate<initon_fraction; %converts from double to logical

%store kon in case there is a delay
kon = rxn_params.k(:,7);

%initialize state structure
%static
init_state = struct;
init_state.t = zeros(nCells,1); 
init_state.tindex = ones(nCells,1);
init_state.x = zeros(nCells,ngenes+1);
init_state.rnase = zeros(nCells,ngenes+1);

n_init_rna = ceil(max(rxn_params.k(:,1) ./ rxn_params.k(:,4) .* ...
    rxn_params.k(:,7) ./ (rxn_params.k(:,7) + rxn_params.k(:,8)))) + 90;
if ~isfinite(n_init_rna)
    n_init_rna = 1;
end
% n_init_rna = 1;

%dynamic: preallocate max(ki/kd) + 90 molecules
init_state.RNA_53_indices = zeros(nCells,n_init_rna,2,'uint16');
init_state.RNA_rnase_binding = false(nCells,n_init_rna,ngenes+1);
init_state.geneactivestate = geneactivestate;
init_state.RNA_identity = false(nCells,n_init_rna,ngenes+1);
init_state.fill = false(nCells,n_init_rna);


if strcmp(experiment_mode,'on')
else
%equilibration is necessary, but no storage
    save_mode = 'markov';
    mode_data = struct; mode_data.tend = Teq;
    [fin_state] = gill_timestepping(init_state,rxn_params,save_mode,mode_data);
    init_state = fin_state;
end

%set time to beginning of domain
init_state.t = zeros(nCells,1); 
init_state.tindex = ones(nCells,1);

if delay>0
%a delay needs to be computed
    delay_tindex = sum(time_array(1,:) <= delay,2);
    
    if strcmp(experiment_mode,'on')
        rxn_params.k(:,7) = 0; %no initiation at the beginning
    end
    
    save_mode = 'save';
    
    mode_data = struct; 
    mode_data.tvec = time_array;
    mode_data.tend = delay;
    mode_data.nascentprobemesh = nascentprobemesh; 
    mode_data.matureprobemesh = matureprobemesh; 
    mode_data.debug_mesh = debug_mesh; 
    mode_data.movie_struct = movie_struct; 
    mode_data.animation_mode = animation_mode;
    
    [fin_state,probemeshes,movie_struct] = gill_timestepping(...
        init_state,rxn_params,save_mode,mode_data);
    
    t = delay*ones(nCells,1);
    tindex = (delay_tindex)*ones(nCells,1);
    
    init_state = fin_state;
    init_state.t = t; init_state.tindex = tindex; 
    
    nascentprobemesh = probemeshes.nascentprobemesh; 
    matureprobemesh = probemeshes.matureprobemesh;
    debug_mesh = probemeshes.debug_mesh;
end

if strcmp(experiment_mode,'on')
    rxn_params.k(:,7) = kon;
elseif strcmp(experiment_mode,'off')
    rxn_params.k(:,7) = 0;
elseif strcmp(experiment_mode,'ss')
end

%normal mode
save_mode = 'save';

% rxn_params.k = k;
% rxn_params.ngenes = ngenes;

mode_data = struct; 
mode_data.tvec = time_array;
mode_data.nascentprobemesh = nascentprobemesh; 
mode_data.matureprobemesh = matureprobemesh; 
mode_data.debug_mesh = debug_mesh;
mode_data.movie_struct = movie_struct;
mode_data.animation_mode = animation_mode;

[~,probemeshes,movie_struct] = gill_timestepping(init_state,rxn_params,save_mode,mode_data);
nascentprobemesh = probemeshes.nascentprobemesh; 
matureprobemesh = probemeshes.matureprobemesh; 
debug_mesh = probemeshes.debug_mesh;
return

%gill_timestepping executes the Gillespie algorithm loop, running
%reactions until it reaches the final requested time point. If appropriate,
%it saves the results to an array.
function [state,probemeshes,movie_struct] = gill_timestepping(init_state,rxn_params,save_mode,mode_data)
ngenes = rxn_params.ngenes;
mult_conds = size(rxn_params.k,1)>1;

num_probes = rxn_params.num_probes; nCells = rxn_params.nCells;

state_saved = struct;
state_saved.x = NaN(nCells,ngenes+1);
state_saved.rnase = NaN(nCells,ngenes+1);

%initialize the structure that saves the end state
k = rxn_params.k;
n_init_rna = ceil(max(k(:,1) ./ k(:,4) .* k(:,7) ./ (k(:,7) + k(:,8)))) + 90;
if ~isfinite(n_init_rna)
    n_init_rna = 1;
end

state_saved.RNA_53_indices = zeros(nCells,n_init_rna,2,'uint16');
state_saved.RNA_rnase_binding = false(nCells,n_init_rna,ngenes+1);
state_saved.geneactivestate = false(nCells,ngenes);
state_saved.RNA_identity = false(nCells,n_init_rna,ngenes+1);
state_saved.fill = false(nCells,n_init_rna);

simindices = 1:nCells;
    
if strcmp(save_mode,'save')
    t_matrix = mode_data.tvec;
    nascentprobemesh = mode_data.nascentprobemesh;
    matureprobemesh = mode_data.matureprobemesh;
    debug_mesh = mode_data.debug_mesh;
    movie_struct = mode_data.movie_struct;

    if isfield(mode_data,'tend')
%This is not ideally tested
        tend = mode_data.tend;
        num_t_pts = sum(t_matrix(1,:)<=tend,2);
    else
        num_t_pts = size(t_matrix,2);
    end
else
    t_matrix = [zeros(size(mode_data.tend)) mode_data.tend];
    num_t_pts = 2;
end

% tend_vec = t_matrix(:,end);

HANDLESTATE = HandleState_int(...
    init_state.RNA_53_indices,...
    init_state.RNA_rnase_binding,...
    init_state.RNA_identity,...
    init_state.geneactivestate,...
    init_state.fill); 

state = struct;
state.t = init_state.t;
state.tindex = init_state.tindex;
state.x = init_state.x;
state.rnase = init_state.rnase;

clearvars init_state;

activecells = true(nCells,1);

while any(activecells)
%     state.t
    torig = state.t; 
    tindex_orig = state.tindex;
    
    mu = zeros(nCells,1);
    
    [t_total,mu_total] = rxn_calculator(...
        state.x(activecells,:),...
        state.rnase(activecells,:),...
        state.t(activecells),...
        HANDLESTATE.gs(activecells,:),...
        rxn_params.k(activecells,:),...
        ngenes);

    state.t(activecells) = t_total;
    mu(activecells) = mu_total;
    
    
    if ~mult_conds
        update = state.t(activecells)>t_matrix(state.tindex(activecells))';
    else
        linindupdate = sub2ind(size(t_matrix),(1:length(state.tindex(activecells)))',...
            state.tindex(activecells));
        tvec_time = t_matrix(linindupdate);
        update = false(nCells,1);
        update(activecells) = state.t(activecells)>tvec_time;
    end
    
    while any(update)
        tobeupdated = find(update);
        
        %Store the probe values if necessary
        if strcmp(save_mode,'save')
            
        [nascentprobestate_previous,matureprobestate_previous] = ...
            probe_calculator(state.x(update,:),...
            rxn_params.num_probes,rxn_params.probes,ngenes,...
            HANDLESTATE.ind(update,:,:),HANDLESTATE.id(update,:,:));
    
            for i = 1:length(tobeupdated)
                %The debug mesh stores the number of molecules of each
                %fluorescently distinct species
                debug_mesh(simindices(tobeupdated(i)),state.tindex(tobeupdated(i)),:) = ...
                    (state.x(tobeupdated(i),:));
                
                for j = 1:num_probes
                    nascentprobemesh(simindices(tobeupdated(i)),state.tindex(tobeupdated(i)),j,:) = ...
                        nascentprobestate_previous(i,j,:);
                    matureprobemesh(simindices(tobeupdated(i)),state.tindex(tobeupdated(i)),j) = ...
                        matureprobestate_previous(i,j);
                end
                
                if mode_data.animation_mode && simindices(tobeupdated(i))==1
                    movie_struct.RNA_53_indices{state.tindex(tobeupdated(i))} = ...
                        HANDLESTATE.ind(tobeupdated(i),:,:);
                    movie_struct.RNA_identity{state.tindex(tobeupdated(i))} = ...
                        HANDLESTATE.id(tobeupdated(i),:,:);
                    movie_struct.RNA_rnase_binding{state.tindex(tobeupdated(i))} = ...
                        HANDLESTATE.bnd(tobeupdated(i),:,:);
                    movie_struct.fill{state.tindex(tobeupdated(i))} = ...
                        HANDLESTATE.fill(tobeupdated(i),:,:);
                    movie_struct.gene_state{state.tindex(tobeupdated(i))} = ...
                        HANDLESTATE.gs(tobeupdated(i),:);
                    movie_struct.x{state.tindex(tobeupdated(i))} = ...
                        state.x(tobeupdated(i),:);
                    movie_struct.rnase{state.tindex(tobeupdated(i))} = ...
                        state.rnase(tobeupdated(i),:);
                end
            end
        end
        
        state.tindex = state.tindex+update;
        ended_in_update = state.tindex(update)>num_t_pts;

        %If any simulations have ended, their state needs to be stored and
        %they need to be marked as inactive
        %Stop state from being saved
        if any(ended_in_update)
            ended = tobeupdated(ended_in_update);
            endedindices = simindices(ended);    
            
            state_saved.x(endedindices,:) = state.x(ended,:);
            state_saved.rnase(endedindices,:) = state.rnase(ended,:);
            
            activecells(ended) = false;
            mu(ended) = 0;
            
            %adjust storage array if necessary
            num_nas_rna = size(HANDLESTATE.ind,2);
            curr_save_num_nas_rna = size(state_saved.RNA_rnase_binding,2);

            if curr_save_num_nas_rna<num_nas_rna
                state_saved.RNA_53_indices = ...
                    [state_saved.RNA_53_indices NaN(nCells,...
                    num_nas_rna-curr_save_num_nas_rna,2)];
                state_saved.RNA_rnase_binding = ...
                    [state_saved.RNA_rnase_binding false(nCells,...
                    num_nas_rna-curr_save_num_nas_rna,ngenes+1)];
                state_saved.RNA_identity = ...
                    [state_saved.RNA_identity false(nCells,...
                    num_nas_rna-curr_save_num_nas_rna,ngenes+1)];
                state_saved.fill = ...
                    [state_saved.fill false(nCells,...
                    num_nas_rna-curr_save_num_nas_rna)];
            end

            %Fill storage array
            state_saved.RNA_53_indices(endedindices,1:num_nas_rna,:) = ...
                HANDLESTATE.ind(ended,1:num_nas_rna,:);
            state_saved.RNA_rnase_binding(endedindices,1:num_nas_rna,:) = ...
                HANDLESTATE.bnd(ended,1:num_nas_rna,:);
            state_saved.RNA_identity(endedindices,1:num_nas_rna,:) = ...
                HANDLESTATE.id(ended,1:num_nas_rna,:);
            state_saved.fill(endedindices,1:num_nas_rna) = ...
                HANDLESTATE.fill(ended,1:num_nas_rna);

            state_saved.geneactivestate(endedindices,:) = HANDLESTATE.gs(ended,:);
            
            
            if ~any(activecells),break;end
        end
        
        %checks if any simulations have reached the next point that must be
        %saved
        if ~mult_conds
            update = state.t(activecells)>t_matrix(state.tindex(activecells))';
        else
            linindupdate = sub2ind(size(t_matrix),(1:length(state.tindex(activecells)))',...
                state.tindex(activecells));
            tvec_time = t_matrix(linindupdate);
            update = false(nCells,1);
            update(activecells) = state.t(activecells)>tvec_time;
        end

    end
    
    %update state
    [state,HANDLESTATE,rollback] = markovstep(state,rxn_params,HANDLESTATE,mu);

    %if reactions are rejected, need to return to previous time
    if ~isempty(rollback)
        state.t(rollback) = torig(rollback);
        state.tindex(rollback) = tindex_orig(rollback);
    end
end

%Check whether RNA were incorrectly accounted for
for space = 1:(ngenes+1)
    if any(sum(state_saved.RNA_identity(:,:,space),2) ~= state_saved.x(:,space))
        error('Saved miscounted RNA in chemspace %i!',space);
    end
end

state = state_saved;

%Check whether probe arrays are fully filled
probemeshes = struct;
if strcmp(save_mode,'save') 
    if any(isnan(nascentprobemesh(:))) && ~isfield(mode_data,'tend')
        warning('nas probe mesh not filled');
    end
    if any(isnan(matureprobemesh(:))) && ~isfield(mode_data,'tend')
        warning('cyt probe mesh not filled');
    end
    probemeshes.nascentprobemesh = nascentprobemesh;
    probemeshes.matureprobemesh = matureprobemesh;
    probemeshes.debug_mesh = debug_mesh;
end
return

function [probe_nas_results,probe_mat_results] = ...
    probe_calculator(x,num_probes,probes,ngenes,...
    RNA_53_indices,RNA_identity)

niter = size(x,1);
chemspace_probe = NaN(niter,num_probes,ngenes+1);
for i = 1:num_probes
    PROBE5 = probes(1,i); PROBE3 = probes(2,i);
    LEFT = max(PROBE5,RNA_53_indices(:,:,2));
    RIGHT = min(PROBE3,RNA_53_indices(:,:,1));
    LENGTH = RIGHT-LEFT;
    contrib = double((RNA_53_indices(:,:,2) < PROBE3) ...
        & (RNA_53_indices(:,:,1) > PROBE5)) ...
        .* double(LENGTH);
     
    for chem_space = 1:(ngenes+1)
        chem_space_stencil = contrib.*double(RNA_identity(:,:,chem_space));
        chem_space_contribution = sum(chem_space_stencil,2,'omitnan');
        chemspace_probe(:,i,chem_space) = chem_space_contribution;
    end
end
probe_nas_results = chemspace_probe(:,:,1:ngenes);
probe_mat_results = chemspace_probe(:,:,ngenes+1);
return

function [t,mu] = rxn_calculator(x,rnase,t,gs,k,ngenes)
%Compute the reaction propensities and identify reactions to run

nCells = size(t,1);

if any(k(:)<0)
    error('A negative reaction rate has been defined!');
end

gene_prop = NaN(nCells,ngenes*6);

no_rnase = x-rnase;


for gene = 1:ngenes
    nas_with_rnase_gene_sum = rnase(:,gene);
    nas_without_rnase_gene_sum = no_rnase(:,gene);
    
    gene_prop(:,(gene-1)*6+1:gene*6) = ...
        [k(:,1).*gs(:,gene) ...
        k(:,2).*x(:,gene) ...
        k(:,3).*nas_without_rnase_gene_sum ...
        k(:,5).*nas_with_rnase_gene_sum ...
        k(:,7).*~gs(:,gene) ...
        k(:,8).*gs(:,gene)];
end

mat_with_rnase_sum = rnase(:,ngenes+1);

mat_without_rnase_sum = no_rnase(:,ngenes+1);
mat_prop = ...
    [k(:,4).*mat_without_rnase_sum  ...
    k(:,6).*mat_with_rnase_sum];

a = [gene_prop mat_prop];

a0 = sum(a,2);

t = t + log(1./rand(nCells,1))./a0;
r2ao = a0.*rand(nCells,1);
mu = sum(repmat(r2ao,1,size(a,2)+1) >= cumsum([zeros(nCells,1),a],2),2);
return

%markovstep modifies the state of the simulation according to the reaction.
%Due to the complex data structures involved, the reactions are hardcoded.
%Inputs: init_state - static variable structure, passed by value.
%                     Has information about molecule and RNase counts.
%        rxn_params - structure with number of genes, gene length, and
%                     simulation coarseness.
%        HANDLESTATE - dynamic variable structure, passed by reference.
%                      Has information about indices and identities of each
%                      mRNA, as well as gene state.
%        mu - vector of reaction indices that occur at each cell.
%Outputs: fin_state - modified version of init_state, passed by value.
%         HANDLESTATE - modified version of HANDLESTATE input.
%         rollback - vector of cell indices that must be reverted to their 
%                    previous state due to a disallowed reaction.
function [fin_state,HANDLESTATE,rollback] = ...
    markovstep(init_state,rxn_params,HANDLESTATE,mu)
x = init_state.x;
rnase = init_state.rnase;
ngenes = rxn_params.ngenes;
L = rxn_params.gene_length;
Nuc = rxn_params.Nuc;

niter = size(x,1);
no_rnase = x-rnase;

rollback = [];

%%%%%%%%%%%%%%%%%%%
%Reactions that can occur for each gene copy
for gene = 1:ngenes
    %INITIATION
    rxn1bool = mu == 1 + 6*(gene-1);

    if any(rxn1bool)
        rxn1ind = find(rxn1bool);
        nRNA = size(HANDLESTATE.fill,2);

        %Expand dynamic array if necessary
        [fillmaxnas,empnas] = max(~HANDLESTATE.fill(rxn1bool,:),[],2);
        if any(fillmaxnas == 0)
            nRNA = nRNA+1;
            empnas(fillmaxnas == 0) = nRNA;
            HANDLESTATE.ind = [HANDLESTATE.ind NaN(niter,1,2)];
            HANDLESTATE.id = [HANDLESTATE.id false(niter,1,ngenes+1)];
            HANDLESTATE.bnd = [HANDLESTATE.bnd false(niter,1,ngenes+1)];
            HANDLESTATE.fill = [HANDLESTATE.fill false(niter,1)];
        end
        
        %Fill in dynamic arrays at appropriate index
        linearindrxn1 = sub2ind([niter,nRNA],rxn1ind,empnas);
        HANDLESTATE.ind([linearindrxn1;linearindrxn1+niter*nRNA]) = 0;
        HANDLESTATE.id(linearindrxn1 + niter*nRNA*(gene-1)) = true;
        HANDLESTATE.fill(linearindrxn1) = true;

        %Update static arrays
        x(:,gene) = x(:,gene) + rxn1bool;
        no_rnase(:,gene) = no_rnase(:,gene) + rxn1bool;
    end

    %%%%%%%%%%%%%%%%%%%
    %NASCENT ELONGATION
    rxn2bool = mu == 2 + 6*(gene-1);

    if any(rxn2bool)
        rxn2ind = find(rxn2bool);
        
        [elong_glob_index,nRNA] = ...
            find_glob_index(HANDLESTATE.id(rxn2bool,:,gene),...
            x(rxn2bool,gene));
        
        linearindrxn2 = sub2ind([niter,nRNA],rxn2ind,elong_glob_index);
        HANDLESTATE.ind(linearindrxn2) = HANDLESTATE.ind(linearindrxn2) + Nuc;

        nas_elong_maturation_array = HANDLESTATE.ind(linearindrxn2) >= L;
        nas_elong_maturation_bool = any(nas_elong_maturation_array,2);

        %Check whether the nascent RNA has matured
        if any(nas_elong_maturation_bool)
            linearindmat_gene = linearindrxn2(nas_elong_maturation_bool);
            HANDLESTATE.ind(linearindmat_gene) = L; %force the 3' end to equal length

            %Update identity arrays
            HANDLESTATE.id(linearindmat_gene + (gene-1)*niter*nRNA) = false; %remove from gene
            HANDLESTATE.id(linearindmat_gene + ngenes*nRNA*niter) = true; %add to cytosol
            
            x(rxn2ind(nas_elong_maturation_bool),gene) = x(rxn2ind(nas_elong_maturation_bool),gene) - 1;
            x(rxn2ind(nas_elong_maturation_bool),ngenes+1) = x(rxn2ind(nas_elong_maturation_bool),ngenes+1) + 1;
            
            %Transfer whether or not an RNase is attached
            rnase(rxn2ind(nas_elong_maturation_bool),gene) = ...
                rnase(rxn2ind(nas_elong_maturation_bool),gene) ...
                - double(HANDLESTATE.bnd(linearindmat_gene + (gene-1)*niter*nRNA));
            
            rnase(rxn2ind(nas_elong_maturation_bool),ngenes+1) = ...
                rnase(rxn2ind(nas_elong_maturation_bool),ngenes+1) ...
                + double(HANDLESTATE.bnd(linearindmat_gene + (gene-1)*niter*nRNA));
            
            HANDLESTATE.bnd(linearindmat_gene + ngenes*nRNA*niter) = ...
                HANDLESTATE.bnd(linearindmat_gene + (gene-1)*niter*nRNA); %add to cytosol
            HANDLESTATE.bnd(linearindmat_gene + (gene-1)*niter*nRNA) = false; %remove from gene

        end
    end

    %%%%%%%%%%%%%%%%%%%%%
    %NAS RNASE
    rxn3bool = mu == 3 + 6*(gene-1);

    if any(rxn3bool)
%         disp('rnase attach to nas');
        rxn3ind = find(rxn3bool);

        nas_without_rnase_array = HANDLESTATE.id(rxn3bool,:,gene) & ~HANDLESTATE.bnd(rxn3bool,:,gene);

        [nas_rnaseattach_glob_index,nRNA] = ...
            find_glob_index(nas_without_rnase_array,...
            no_rnase(rxn3bool,gene));
        
        %Update dynamic and static arrays
        linearindrxn3 = sub2ind([niter,nRNA],rxn3ind,nas_rnaseattach_glob_index);
        HANDLESTATE.bnd(linearindrxn3 + (gene-1)*niter*nRNA) = true;
        
        rnase(:,gene) = rnase(:,gene) + rxn3bool;
    end

    %%%%%%%%%%%%%%%%%%%%%
    %NASCENT DEGRADATION
    rxn4bool = mu == 4 + 6*(gene-1);

    if any(rxn4bool)
        rxn4ind = find(rxn4bool);

        [nas_deg_glob_index,nRNA] = ...
            find_glob_index(HANDLESTATE.bnd(rxn4bool,:,gene),rnase(rxn4bool,gene));

        linearindrxn4 = sub2ind([niter,nRNA],rxn4ind,nas_deg_glob_index);
        HANDLESTATE.ind(linearindrxn4 + nRNA*niter) = HANDLESTATE.ind(linearindrxn4 + nRNA*niter) + Nuc;
        
        %Check whether reaction is nonphysical, degrading pas the RNAP
        nas_deg_overtake_array = HANDLESTATE.ind(linearindrxn4 + nRNA*niter) > HANDLESTATE.ind(linearindrxn4);
        nas_deg_overtake_bool = any(nas_deg_overtake_array,2);

        if any(nas_deg_overtake_bool)
            linearindovertake_gene = linearindrxn4(nas_deg_overtake_bool);
            rollback = [rollback;rxn4ind(nas_deg_overtake_bool)];
            
            HANDLESTATE.ind(linearindovertake_gene + nRNA*niter) = HANDLESTATE.ind(linearindovertake_gene + nRNA*niter) - Nuc;
        end
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %GENE TURNON
    rxn5bool = mu == 5 + 6*(gene-1);

    if any(rxn5bool)
        HANDLESTATE.gs(rxn5bool,gene) = true;
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%
    %GENE TURNOFF
    rxn6bool = mu == 6 + 6*(gene-1);

    if any(rxn6bool)
        HANDLESTATE.gs(rxn6bool,gene) = false;
    end

end

%%%%%%%%%%%%%%%%%%%%%
%CYTOSOLIC RNASE ATTACHMENT
rxn7bool = mu == 6*ngenes + 1;

if any(rxn7bool)
    rxn7ind = find(rxn7bool);
    mat_without_rnase_array = HANDLESTATE.id(rxn7bool,:,ngenes+1) & ~HANDLESTATE.bnd(rxn7bool,:,ngenes+1);

    [mat_rnaseattach_glob_index,nRNA] = ...
        find_glob_index(mat_without_rnase_array,...
        no_rnase(rxn7bool,ngenes+1));

    linearindrxn7 = sub2ind([niter,nRNA],rxn7ind,mat_rnaseattach_glob_index);
    HANDLESTATE.bnd(linearindrxn7 + ngenes*nRNA*niter) = true;
    
    rnase(:,ngenes+1) = rnase(:,ngenes+1) + rxn7bool;
end

%%%%%%%%%%%%%%%%%%%%%
%CYTOSOLIC DEGRADATION
rxn8bool = mu == 6*ngenes + 2;

if any(rxn8bool)
    rxn8ind = find(rxn8bool);
    [mat_deg_glob_index,nRNA] = ...
        find_glob_index(HANDLESTATE.bnd(rxn8bool,:,ngenes+1),...
        rnase(rxn8bool,ngenes+1));

    linearindrxn8 = sub2ind([niter,nRNA],rxn8ind,mat_deg_glob_index);
    HANDLESTATE.ind(linearindrxn8 + niter*nRNA) = HANDLESTATE.ind(linearindrxn8 + niter*nRNA) + Nuc;

    %Check mature extinction condition
    cyt_deg_overtake_array = HANDLESTATE.ind(linearindrxn8 + niter*nRNA) > HANDLESTATE.ind(linearindrxn8);
    cyt_deg_overtake_bool = any(cyt_deg_overtake_array,2);

    if any(cyt_deg_overtake_bool)
%         disp('CYT EXTINCTION');            
        linearind_extinction = linearindrxn8(cyt_deg_overtake_bool); 
        
        HANDLESTATE.ind(linearind_extinction) = NaN;
        HANDLESTATE.ind(linearind_extinction + nRNA*niter) = NaN;

        HANDLESTATE.fill(linearind_extinction) = false;
        
        HANDLESTATE.id(linearind_extinction + ngenes*nRNA*niter) = false; %remove from chem space

        
        x(rxn8ind(cyt_deg_overtake_bool),ngenes+1) = ...
            x(rxn8ind(cyt_deg_overtake_bool),ngenes+1) - 1;
        
        rnase(rxn8ind(cyt_deg_overtake_bool),ngenes+1) = ...
            rnase(rxn8ind(cyt_deg_overtake_bool),ngenes+1) ...
            - double(HANDLESTATE.bnd(linearind_extinction + ngenes*nRNA*niter));
        HANDLESTATE.bnd(linearind_extinction + ngenes*nRNA*niter) = false;
        
    end
        
end


fin_state = struct;
fin_state.x = x;
fin_state.rnase = rnase;
fin_state.t = init_state.t;
fin_state.tindex = init_state.tindex;
return

%find_glob_index finds the dynamic array index of mRNA to elongate/degrade/
%attach an RNase.
%This function is computationally intensive due to the large number of
%these reactions.
%The function presupposes that each mRNA is equally likely to undergo the
%reaction; more sophisiticated weighing can be implemented.
%Inputs: target_array - 2D binary array with at least one TRUE in every
%                       row.
%        target_array_sum - vector that gives number of TRUE values in each
%                       row of target_array.
%Outputs: global_index - vertical vector that gives the column number of a
%                        random TRUE value from each row of target_array.
%         nRNA - width of target_array.
function [global_index,nRNA] = find_glob_index(target_array,target_array_sum)
num_cells = length(target_array_sum);
local_index = ceil(rand(num_cells,1).*target_array_sum);
nRNA = size(target_array,2);

global_index = sum(cumsum(target_array,2) < local_index,2)+1;
return
